from argparse import ArgumentParser
from .devserver import FileWatcher
from watchdog.observers import Observer
from rich.logging import RichHandler
from pathlib import Path
from .config import Config
import logging


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("--project_root", "-p", default=".")
    parser.add_argument("--verbose", "-v", action="store_true")
    subparser = parser.add_subparsers(dest="command")
    # dev
    dev_parser = subparser.add_parser("dev")
    dev_parser.add_argument("--port", "-P", default=8000)
    dev_parser.add_argument("--addr", "-A", default="0.0.0.0")
    # build
    build_parser = subparser.add_parser("build")

    return parser.parse_args()


async def _main(args):
    project_root = Path(args.project_root)
    if args.verbose:
        logging.basicConfig(handlers=[RichHandler()], level=logging.DEBUG)
    else:
        logging.basicConfig(handlers=[RichHandler()], level=logging.INFO)
    logging.debug(f"Project Root: {args.project_root}")
    if args.command == "dev":
        dev_server = FileWatcher(
            project_root=project_root,
            port=args.port,
            addr=args.addr,
        )
        observer = Observer()
        observer.schedule(dev_server, path=args.project_root, recursive=True)
        observer.start()
        try:
            await dev_server._ws_server._server_task
        except KeyboardInterrupt:
            observer.stop()
            dev_server.terminate()
        observer.join()
    if args.command == "build":
        from arisu.build import ProjectRenderer
        renderer = ProjectRenderer(project_root)
        renderer.build()


def main():
    import asyncio as aio
    aio.run(_main(parse_args()))


if __name__ == "__main__":
    main()


from .template import HTMLRenderer
from typing import Type


class ConfigOverrider:

    def __init__(self):
        self._markdown_render = HTMLRenderer

    def set_markdown_render(self, renderer_cls: Type[HTMLRenderer]):
        if not issubclass(renderer_cls, HTMLRenderer):
            raise TypeError(
                f"{renderer_cls} is not a subclass of HTMLRenderer")
        self._markdown_render = renderer_cls



import logging
from functools import wraps
from typing import Any, Concatenate, ParamSpec, TypeVar, Callable
from pathlib import Path

P = ParamSpec("P")
T = TypeVar("T")
F = TypeVar("F")


def copy_method_signature(
    source: Callable[Concatenate[Any, P], T]
) -> Callable[[Callable[..., T]], Callable[Concatenate[Any, P], T]]:

    def wrapper(target: Callable[..., T]) -> Callable[Concatenate[Any, P], T]:

        @wraps(source)
        def wrapped(self: Any, /, *args: P.args, **kwargs: P.kwargs) -> T:
            return target(self, *args, **kwargs)

        return wrapped

    return wrapper


class LoggerMixin(object):

    def __init__(self, name=__file__):
        self._logger = logging.getLogger(name)

    @copy_method_signature(logging.info)
    def info(self, *args, **kwargs):
        self._logger.info(*args, **kwargs)

    @copy_method_signature(logging.error)
    def error(self, *args, **kwargs):
        self._logger.error(*args, **kwargs)

    @copy_method_signature(logging.warning)
    def warning(self, *args, **kwargs):
        self._logger.warning(*args, **kwargs)

    @copy_method_signature(logging.debug)
    def debug(self, *args, **kwargs):
        self._logger.debug(*args, **kwargs)


def build_keypath_from_relative_path(path: Path):
    p = Path(path)
    suffixes = p.suffixes
    full_path = p.as_posix()
    for suffix in suffixes[::-1]:
        full_path = full_path.removesuffix(suffix)
    if full_path == ".":
        full_path = "index"

    full_path = full_path.replace('/', '.')
    return full_path



import os
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any

from tomli import load

from benedict import benedict

from ..utils import LoggerMixin


def get_dict_val_by_level(val: dict, path: str):
    parts = path.split('.')
    res = val
    for part in parts:
        res = res[part]
    return res


class _ContentUrlIndexer(object):

    def __init__(self, data: benedict | None):
        self._data = data

    def __getitem__(self, item: str):
        path = self._convert_url_to_dotpath(item)
        res = self._data[path]
        res["__globals__"] = self._data.get("__globals__", {})
        return res

    def get(self, key, default=None):
        path = self._convert_url_to_dotpath(key)
        res = self._data.get(path, default)
        res["__globals__"] = self._data.get("__globals__", {})
        return res

    @staticmethod
    def _convert_url_to_dotpath(path: str):
        if path == "/":
            return "index"
        return path.strip('/').replace('/', '.')


class ContentLoader(metaclass=ABCMeta):

    def __init__(self, p: str | Path | Any):
        self._inner: benedict | None = None
        self._content_cfg = p
        self._url_indexer = _ContentUrlIndexer(self._inner)

    @abstractmethod
    def load(self):
        raise NotImplementedError()

    def __getitem__(self, item: str):
        result = None

        try:
            result = self._inner[item]
            result["__globals__"] = self._inner.get("__globals__", {})
            return result
        except KeyError:
            if item.endswith('.index'):
                result = self._inner[item.removesuffix('.index')]
                result["__globals__"] = self._inner.get("__globals__", {})
                return result

    def get(self, key, default=None):
        self._inner.get(key, default)

    @property
    def url_indexer(self) -> _ContentUrlIndexer:
        return self._url_indexer


class ContentPathLoader(ContentLoader, LoggerMixin):

    def __init__(self, p):
        ContentLoader.__init__(self, p)
        LoggerMixin.__init__(self, "ContentDirLoader")
        self.load()

    def load(self):
        p = Path(self._content_cfg) if not isinstance(
            self._content_cfg, Path) else self._content_cfg
        if p.is_file() and p.suffix == ".toml":
            with open(p, 'rb') as f:
                self._inner = benedict(load(f))
        elif p.is_dir():
            self._recursive_load(p)
        self._url_indexer = _ContentUrlIndexer(self._inner)

    def _recursive_load(self, path: Path):
        queue = [path]
        res = benedict()
        while len(queue) > 0:
            print(queue)
            p = queue.pop()
            content = None
            if p.is_file() and p.name.endswith(".toml"):
                with open(p, 'rb') as f:
                    content = benedict(load(f))
            elif p.is_dir():
                queue.extend(list(p.iterdir()))
            if content is not None:
                suffixes = p.suffixes
                full_path = p.relative_to(path).as_posix()
                for suffix in suffixes:
                    full_path = full_path.removesuffix(suffix)
                if full_path == ".":
                    full_path = "index"

                full_path = full_path.replace('/', '.')
                self.debug(f"Adding {full_path} to {content}")
                if full_path in res:
                    raise KeyError(f"Duplicate key {full_path}")
                res[full_path] = content
        self._inner = res


class ContentMappingLoader(ContentLoader):

    def __init__(self, p: Any):
        ContentLoader.__init__(self, p)
        self.load()

    def load(self):
        for cfg in self._content_cfg:
            keypath = self.url_to_benedict_keypath(cfg.name)
            self._inner[keypath] = cfg.content
        self._url_indexer = _ContentUrlIndexer(self._inner)

    @staticmethod
    def url_to_benedict_keypath(key: str):
        if key == "/":
            return "index"
        else:
            return key.strip('/').replace('/', '.')


from types import FunctionType
from flask.templating import Environment
from watchdog.events import FileSystemEventHandler

from .extensions.template import AutoRefreshListener, MarkdownRender

from .extensions import ConfigOverrider
from .config import Config
from flask import Flask, render_template
import multiprocess as mp
from pathlib import Path
from websockets.server import serve, WebSocketServerProtocol
from websockets import broadcast
import asyncio as aio
from .utils import LoggerMixin
import typing as t
import os


class DevEnvironment(Environment):

    def __init__(self, *args, **kwargs):
        post_init_callback = kwargs.pop('post_init_callback', None)
        super().__init__(*args, **kwargs)
        if post_init_callback is not None:
            post_init_callback(self)


class WebSocketNotifier(LoggerMixin):

    def __init__(self, port, addr="0.0.0.0"):
        super().__init__(name="sitgen:WebSocketNotifier")
        self._server = serve(self._register_conn, addr, port)
        self._addr = addr
        self._port = port
        self._conns = set()

        self._queue = aio.Queue()
        self._server_task = None

    async def _register_conn(self, websocket: WebSocketServerProtocol):
        self._conns.add(websocket)
        try:
            await websocket.wait_closed()
        finally:
            self._conns.remove(websocket)

    async def _init_server(self):
        async with self._server:
            print(
                f"WebSocketNotifier listening on ws://{self._addr}:{self._port}"
            )

            await aio.Future()

    def run_in_background(self):
        if self._server_task is None:
            self._server_task = aio.create_task(self._init_server())

        else:
            raise RuntimeError("Server already running")

    async def run(self):
        if self._server_task is None:
            self.run_in_background()
            await self._server_task
        else:
            raise RuntimeError("Server already running")

    def terminate(self):
        if self._server_task is not None:
            self._server_task.cancel()
            self._server_task = None
        self._server.shutdown()

    def send_update_notification(self):
        self.debug("Sending update notification")
        broadcast(self._conns, "RELOAD")


class FileWatcher(FileSystemEventHandler, LoggerMixin):

    def __init__(
            self,
            project_root=Path("."),
            port=8000,
            addr="127.0.0.1",
            ws_port=8088,
    ):

        super().__init__(name="sitgen:ProjectWatcher")
        self._port = port
        self._addr = addr
        self._project_root = project_root if isinstance(
            project_root, Path) else Path(project_root)
        self._server = DevServer(__name__, project_root=self._project_root)
        self._ws_server = WebSocketNotifier(ws_port)
        self._ws_server.run_in_background()
        self.start_server()

    def terminate(self):
        self._ws_server.terminate()
        self.terminate_server()

    def _restart_server(self):
        if self._process is not None:
            self.info("Restarting dev server")
            self.terminate_server()
            self._server = DevServer(__name__, project_root=self._project_root)
        self.start_server()
        if self._process is not None:
            self._ws_server.send_update_notification()

    def start_server(self):
        self._process = self._create_process()
        self._process.start()
        self.info(f"Dev server istening on http://{self._addr}:{self._port}")

    def _create_process(self):
        return mp.Process(
            target=self._server.run,
            kwargs={
                "host": self._addr,
                "port": self._port,
                "debug": False,
            },
        )

    def terminate_server(self):
        if self._process is not None:
            self._process.terminate()
            self._process.join()
            self._process = None

    def on_moved(self, event):
        self._restart_server()

    def on_deleted(self, event):
        self._restart_server()

    def on_modified(self, event):
        self._restart_server()

    def on_created(self, event):
        self._restart_server()


class DevServer(Flask, LoggerMixin):
    jinja_environment = DevEnvironment

    def __init__(
        self,
        import_name: str,
        static_url_path: t.Optional[str] = None,
        static_host: t.Optional[str] = None,
        host_matching: bool = False,
        subdomain_matching: bool = False,
        project_root: t.Optional[t.Union[str, os.PathLike]] = Path("."),
        instance_path: t.Optional[str] = None,
        instance_relative_config: bool = False,
        root_path: t.Optional[str] = None,
    ):
        LoggerMixin.__init__(self, name="sitegen:DevServer")
        self.jinja_options['extensions'] = [
            MarkdownRender, AutoRefreshListener
        ]
        self._project_root = Path(project_root) if not isinstance(
            project_root, str) else project_root
        self._logger.debug(f"Using project root: {self._project_root}")
        self._project_config = Config.from_toml(self._project_root /
                                                "config.toml")
        static_dir = self._project_root / Path(
            self._project_config.general.static_dir)
        template_dir = self._project_root / Path(
            self._project_config.general.template_dir)

        super().__init__(import_name, static_url_path,
                         static_dir.absolute().as_posix(), static_host,
                         host_matching, subdomain_matching,
                         template_dir.absolute().as_posix(), instance_path,
                         instance_relative_config, root_path)

        self._content_loader = self._project_config.general.content_loader(
            Path(self._project_root))
        self._content_loader.load()
        self.derive_routes_from_dir()
        self._merge_overrides()

    def restart(self):
        self.content_loader.load()
        self.derive_routes_from_dir()
        self._merge_overrides()

    def reload_template_content(self):
        from tomli import load
        self.info(f"Loading template content from '{self.content_file}'")
        with open(self.content_file, "rb") as f:
            self._template_content = load(f)

    @property
    def content_loader(self):
        return self._content_loader

    def derive_routes_from_dir(self):
        for path in Path(self.template_folder).iterdir():
            if path.is_dir():
                self.derive_routes_from_dir()
            elif path.as_posix().endswith(
                    ".html.jinja") and not path.stem.startswith("_"):
                url_path = "/" if path.stem.startswith(
                    "index.") else f"/{path.stem.split('.')[0]}"
                context = self.content_loader.url_indexer.get(url_path, {})
                context["SITEGEN_ENV"] = "dev"
                self.add_url_rule(
                    url_path,
                    endpoint=url_path,
                    view_func=self._create_view_func(
                        url_path.replace("/", "_"),
                        path.relative_to(self.template_folder).as_posix(),
                        **context,
                    ),
                )
                self.info(f"Added route '{url_path}' -> '{path}'")
                self._logger.debug(f"route '{url_path}' context: {context}")

    def _set_jinja_callback(self, overrides: ConfigOverrider):

        def callback(env: DevEnvironment):
            env.markdown_render = overrides._markdown_render

        self.jinja_options['post_init_callback'] = callback

    def _merge_overrides(self):
        import sys
        import importlib
        sys.path.insert(1, self._project_root.absolute().as_posix())
        mod = importlib.import_module("overrides")
        self._set_jinja_callback(mod.sitegen_overrides(ConfigOverrider()))

    @staticmethod
    def _create_view_func(prefix, path: str, **context):

        def view_func():
            return render_template(path, **context)

        fn = FunctionType(
            view_func.__code__,
            view_func.__globals__,
            f"{prefix}_view_func",
            view_func.__defaults__,
            view_func.__closure__,
        )
        return fn



from pathlib import Path
from pydantic import BaseModel
from os import PathLike

from .loaders.content import ContentMappingLoader, ContentPathLoader


class ContentMapping(BaseModel):
    content: str
    name: str


class GeneralConfig(BaseModel):
    template_dir: str = "templates"
    output_dir: str = "output"
    static_dir: str = "static"
    content: str | ContentMapping = "content.toml"

    def content_loader(self, project_root=Path('.')):
        if isinstance(self.content, str):
            p = project_root / Path(self.content)
            is_toml = p.suffix == ".toml"
            is_dir = p.is_dir()
            if is_dir or is_toml:
                return ContentPathLoader(p)
            else:
                raise ValueError(
                    f"Invalid content config: {self.content} is not a valid path"
                )
        elif isinstance(self.content, ContentMapping):
            return ContentMappingLoader(self.content)
        else:
            raise ValueError(
                f"Invalid content config: {self.content} is not a valid config"
            )


class DeployConfig(BaseModel):
    branch: str = "gh-pages"


class Route(BaseModel):
    path: str
    template: str

    @property
    def name(self):
        return self.path[1:]


class RouterConfig(BaseModel):
    routes: list[Route] = []
    hash_route = False


class Config(BaseModel):
    general: GeneralConfig
    deploy: DeployConfig

    @classmethod
    def from_toml(cls, path: "PathLike") -> "Config":
        from tomli import load
        with open(path, "rb") as f:
            return cls(**load(f))



from pathlib import Path


class Render:

    def __init__(self, projecct_dir: str):
        pass

    @property
    def output_dir(self):
        return Path(self._conf.output_dir)

    @property
    def static_dir(self):
        return Path(self._conf.output_dir)

    @property
    def template_dir(self):
        return Path(self._conf.template_dir)



from os import PathLike
from .config import Config
from . import ConfigOverrider
from jinja2 import Environment, FileSystemLoader, select_autoescape
from pathlib import Path
from .extensions.template import MarkdownRender, AutoRefreshListener
import sys
import importlib
from .utils import LoggerMixin, build_keypath_from_relative_path
from shutil import copy, copytree


class ProjectRenderer(LoggerMixin):

    def __init__(self, project_dir: PathLike):
        super().__init__(name="sitegen:build")
        self._project_dir = project_dir if isinstance(
            project_dir, Path) else Path(project_dir)
        self._config = Config.from_toml(self._project_dir / 'config.toml')

        self._output_dir = self._project_dir / self._config.general.output_dir
        self._output_dir.mkdir(parents=True, exist_ok=True)
        self._content_loader = self._config.general.content_loader(
            self._project_dir)
        self._template_loader = FileSystemLoader(
            self._project_dir / self._config.general.template_dir)
        self._jinja_env = Environment(
            loader=self._template_loader,
            extensions=[MarkdownRender, AutoRefreshListener],
            autoescape=select_autoescape(['html', 'xml']),
        )
        self._merge_overrides()

    def handle_templates_dir(self):
        for template_path in self._template_loader.list_templates():
            p = Path(template_path)
            self.debug(f"Found file {template_path} in template dir")
            if not p.name.startswith("_") and p.name.endswith(".jinja"):
                self.info(f"Rendering {template_path}")
                template = self._jinja_env.get_template(template_path)
                keypath = build_keypath_from_relative_path(p)
                content = self._content_loader[keypath]
                rendered = template.render(**content)
                output_path = self._output_dir / p
                output_path.parent.mkdir(parents=True, exist_ok=True)
                output_path = output_path.absolute().as_posix().removesuffix(
                    ".jinja")
                self.debug(f"Writing template to {output_path}")
                with open(output_path, 'w') as f:
                    f.write(rendered)
            elif not p.name.startswith("_"):
                src = p.absolute().as_posix()
                dst = (self._output_dir / p).absolute()
                dst.parent.mkdir(parents=True, exist_ok=True)
                self.debug(f"Copying {src} to {dst}")
                copy(src, dst.as_posix())

    def handle_static_dir(self):
        src = self._project_dir / self._config.general.static_dir
        dst = self._output_dir / self._config.general.static_dir
        dst.parent.mkdir(parents=True, exist_ok=True)
        self.debug(f"Copying static dir ('{src.as_posix()}') to {dst}")
        copytree(src.as_posix(), dst.as_posix())

    def _merge_overrides(self):
        sys.path.insert(1, self._project_dir.absolute().as_posix())
        try:
            overrides = importlib.import_module('overrides')
            overrider = ConfigOverrider()
            overrider: 'ConfigOverrider' = overrides.sitegen_overrides(
                overrider)
            self._jinja_env.markdown_render = overrider._markdown_render
        except ModuleNotFoundError:
            self.info(f"No overrides found in {self._project_dir}")

    def build(self):
        self.handle_templates_dir()
        self.handle_static_dir()



from jinja2.ext import Extension
from mistune import HTMLRenderer, create_markdown
from jinja2 import Environment, nodes
from jinja2.parser import Parser

from ..utils import LoggerMixin


class AutoRefreshListener(Extension, LoggerMixin):
    tags = {"autorefresh"}

    def __init__(self, environment: Environment):
        super().__init__(environment)
        LoggerMixin.__init__(self, name="sitegen:AutoRefreshListener")

    def parse(self, parser: Parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        print(args)
        return nodes.CallBlock(self.call_method("_render_autorefresh", args),
                               [], [], []).set_lineno(lineno)

    def _render_autorefresh(self, port, caller):
        return f"""
        <script>
            var socket = new WebSocket("ws://localhost:{port}");
            socket.onmessage = function(event) {{
                console.log(event.data);
                if (event.data == "RELOAD") {{
                    location.reload();
                }}
            }}
            socket.onopen = (() => socket.send("REIGSTER"));
        
        </script>
        """


class MarkdownRender(Extension, LoggerMixin):
    tags = {"markdown"}

    def __init__(self, environment: Environment):
        super().__init__(environment)
        LoggerMixin.__init__(self, name="sitegen:MarkdownRender")
        environment.extend(markdown_render=HTMLRenderer)

    def parse(self, parser: Parser):
        lineno = next(parser.stream).lineno

        body = parser.parse_statements(["name:endmarkdown"], drop_needle=True)

        return nodes.CallBlock(self.call_method("_render_markdown", []), [],
                               [], body).set_lineno(lineno)

    def _render_markdown(self, caller):
        renderer: HTMLRenderer = self.environment.markdown_render()
        markdown = create_markdown(renderer=renderer)

        body = caller()
        html = markdown(body.strip())
        return html



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arisu', 'arisu.extensions', 'arisu.loaders']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.2.3,<3.0.0',
 'jinja2>=3.1.2,<4.0.0',
 'mistune>=2.0.5,<3.0.0',
 'multiprocess>=0.70.14,<0.71.0',
 'pydantic>=1.10.7,<2.0.0',
 'python-benedict>=0.30.0,<0.31.0',
 'rich>=13.3.3,<14.0.0',
 'tomli>=2.0.1,<3.0.0',
 'watchdog>=3.0.0,<4.0.0',
 'websockets>=11.0,<12.0']

entry_points = \
{'console_scripts': ['aris = aris.__main__:main']}

setup_kwargs = {
    'name': 'arisu',
    'version': '0.1.2',
    'description': '',
    'long_description': '<!-- @format -->\n\ngi# sitegen\n',
    'author': 'Alex Zhang',
    'author_email': 'zhangchi0104@live.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
